unit ICHelperWhat;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the Helper What, which is used to drive the entire
	Helper dialog.  The dialog is very similar to the File Mapping dialog,
	and the Whats have a lot in common, including a lot of block copied code.
	Hey, we were being slack!  I've tried to eliminate the grosser duplications
	by moving some of the code into ICWhatSubs, but there's still a lot of
	common code.
	
	The bulk of the code falls into two chunks. One chunk (which comes second
	in the file) operates the Helper window itself, with the usual array of
	standard What functions and commands that implement each of the buttons.
	The other chunk (which comes first in the file) is responsible for driving
	the Add/Edit Helper modal dialog.
	
	The key data structures is a HelperEntry list stored in the whatData field,
	and a ListManager ListHandle in the whatSpareData.  The HelperEntry list
	is a much simpler data structure than the corresponding ICMapEntry list
	in the File Mapping window.  It's basically an unbounded array of fixed
	sized elements that are HelperEntries. Each HelperEntry holds
	the URL scheme for the entry, the ICAppSpec for that scheme, and a locked
	Boolean.  This locked Boolean is weird because its stored on a scheme-by-
	scheme basis rather than for the entire preference set.  This is because
	the list of helpers is actually stored as a bunch of IC preferences,
	each of which may be locked, rather than just one preference (which
	is the case for the File Map Whats).
	
	The ListManger list is only used for display and editing purposes, and no
	data is actually stored in the list.

	The module implements its own 'LDEF' to draw the list. This is done
	using a neat trick, namely the SimpleLDEF. This LDEF is a stub that
	assumes that the list's refCon field points to a procedure that
	implements the real LDEF, and simply routes all LDEF calls through
	to the refCon.
*)

interface

	uses
		ICWindowGlobals;

	(* ***** Standard What Interface ***** *)
	
	(* See the comments in ICWindowGlobals. *)

	function WhatInitHelper(dummy : integer) : OSStatus;
	function WhatOpenHelper (wt: WindowType; item: integer): OSStatus;
	function WhatFlushHelper (wt: WindowType; item: integer): OSStatus;
	function WhatCloseHelper (wt: WindowType; item: integer): OSStatus;
	function WhatKeyHelper (wt: WindowType; item: integer; event: EventRecord): OSStatus;
	function WhatClickHelper (wt: WindowType; item: integer; event: EventRecord): OSStatus;
	function WhatActivateHelper (wt: WindowType; item: integer; activate: Boolean): OSStatus;
	function WhatCursorHelper (wt: WindowType; item: integer; pt: Point; cursorID: integer): OSStatus;

	(* ***** Layer Breaking Interface ***** *)

	procedure HelperSetSelection(key : Str255);
		(* Called by the FindWindowAndSelectKey routine in ICEditPrefAppleEvents.
			This routine is called to select the entry in the Helpers list that
			is best described by the key. In this case the key is the helper name.
			
			For example, if an Edit Preference AppleEvent is sent with
			the parameter "Helperhttp", the FindWindowAndSelectKey matches
			and strips off the "Mapping", and passes "http" in as the key parameter
			to this routine.
		*) 

implementation

	uses
		Icons,
		Dialogs, 
		Lists, 
		Fonts, 
		ToolUtils, 
		Errors, 

		InternetConfig,
		
		ICResources, 
		ICGlobals,
		ICMiscSubs, 
		ICCommonSubs,
		ICWhatSubs,
		ICStringLists, 
		ICDialogs, 
		ICIconSuites, 
		ICStandardFile, 
		ICMovableModal, 
		ICDocument;

	var
		gListUserItemUPP : UserItemUPP;
		gFileMapLDEFUPP : ListDefUPP;
	
	const
		kHelperCellHeight = kListCellHeight;
		(* kHelperCellHeight is the height of the cells in the Helper list. *)

	type
		HelperEntry =
			record
				urlScheme: Str255;
				applicationForScheme: ICAppSpec;
				entryIsLocked: Boolean;
			end;
		(* HelperEntry holds the information about one helper application. *)

	(* ***** Add/Edit Helper Modal Dialog  ***** *)

	procedure FixEditHelperButton (dlg: DialogPtr);
		(* This routine is responsible for dimming/undimming the OK button
			for the Edit Helper modal dialog. It undims the OK button if
			a helper application has been chosen and a URL scheme entered.
			This routine is outside of DoEditHelperModalDialog because it's called from
			EditHelperFilter, which is required to be outside of DoEditHelperModalDialog. 
		*)
		var
			helperApplicationName : Str255;
			urlScheme: Str255;
	begin
		GetDItemText(dlg, ditAddHelperAppName, helperApplicationName);
		GetDItemText(dlg, ditAddHelperScheme, urlScheme);
		SetDControlEnable(dlg, ditOK, (helperApplicationName <> '') and (urlScheme <> ''));
		DefaultButtonUserItem(dlg, ditAddHelperDefaultUserItem);
	end; (* FixEditHelperButton *)

	function EditHelperFilter (dlg: DialogPtr; var event: EventRecord; var item: integer): Boolean;
		(* This routine is the filter function used by the Edit Helper modal
			dialog.  It is outside of DoEditHelperModalDialog because filter functions cannot
			be nested.
		*)
		var
			result: Boolean;
	begin
		result := OKCancelModalFilter(dlg, event, item);
		FixEditHelperButton(dlg);
		EditHelperFilter := result;
	end; (* EditHelperFilter *)

	function DoEditHelperModalDialog (var entryToEdit: HelperEntry): OSStatus;
		(* This routine poses the Edit Helper modal dialog based on the information
			in entryToEdit. It collects the results and returns them also in
			entryToEdit.
		*)

		function ChooseHelperApplication (var chosenApp: ICAppSpec): Boolean;
			(* This routine puts up a Standard File dialog to ask the user
				for a helper application. It returns true if an application
				was successfully chosen, false otherwise.
			*)
			var
				err: OSStatus;
				chosenFileSpec: FSSpec;
				chosenFileInfo: FInfo;
		begin
			err := ICStandardGetFile('APPL', chosenFileSpec, chosenFileInfo);
			DisplayError(acChooseApplication, err);
			if err = noErr then begin
				chosenApp.fCreator := chosenFileInfo.fdCreator;
				chosenApp.name := chosenFileSpec.name;
			end; (* if *)
			ChooseHelperApplication := (err = noErr);
		end; (* ChooseHelperApplication *)

		var
			err: OSStatus;
			editHelperDialog: DialogPtr;
			hitItem: integer;
			movableModalSavedMenuBarState: Ptr;
	begin
		(* Create the dialog. *)
		err := noErr;
		editHelperDialog := GetNewDialog(rEditHelperDialog, nil, WindowPtr(-1));
		if editHelperDialog = nil then begin
			err := memFullErr;
		end; (* if *)
		
		(* Pose the dialog. *)
		if err = noErr then begin
			(* Insert the information from entryToEdit into the fields of the dialog.
				Start off by removing "Helper" from the front of the urlScheme.
			*)
			Delete(entryToEdit.urlScheme, 1, length(kICHelper));
			SetDItemText(editHelperDialog, ditAddHelperScheme, entryToEdit.urlScheme);
			SetDItemText(editHelperDialog, ditAddHelperAppName, entryToEdit.applicationForScheme.name);
			SelectDialogItemText(editHelperDialog, ditAddHelperScheme, 0, 32767);
			SetupDefaultButtonUserItem(editHelperDialog, ditOK, ditAddHelperDefaultUserItem);
			FixEditHelperButton(editHelperDialog);
			
			(* Run the dialog, processing each hitItem. *)
			ShowWindow(editHelperDialog);
			DisableMenuBar(movableModalSavedMenuBarState);
			InitCursor;
			repeat
				MovableModalDialog(@EditHelperFilter, hitItem);
				case hitItem of
					ditAddHelperChooseApplication:  begin
						if ChooseHelperApplication(entryToEdit.applicationForScheme) then begin
							SetDItemText(editHelperDialog, ditAddHelperAppName, entryToEdit.applicationForScheme.name);
							FixEditHelperButton(editHelperDialog);
						end; (* if *)
					end;
					otherwise
						(* do nothing *) ;
				end; (* case *)
			until hitItem in [ditOK, ditCancel];
			
			(* Process the results. If the user clicked OK, extract the information
				from the dialog fields and return it in entryToEdit.
			*)
			if hitItem = ditOK then begin
				GetDItemText(editHelperDialog, ditAddHelperScheme, entryToEdit.urlScheme);
				(* Prepend "Helper" back on the urlScheme to form the correct
					IC key again.
				*)
				entryToEdit.urlScheme := concat(kICHelper, entryToEdit.urlScheme);
			end else begin
				err := userCanceledErr;
			end; (* if *)
			
			(* Clean up. *)
			ReEnableMenuBar(movableModalSavedMenuBarState);
			DisposeDialog(editHelperDialog);
		end; (* if *)
		DoEditHelperModalDialog := err;
	end; (* DoEditHelperModalDialog *)

	(* ***** Utilities for the Helpers Window ***** *)

	(* ***** Accessors for the Primary Data Structure ***** *)

	procedure GetNthElement (entries: Handle; index: integer; var result: HelperEntry);
		(* This routine returns the index'th HelperEntry in the entries handle.
			0 <= index < number of entries
		*)
	begin
		BlockMoveData(Ptr(longint(entries^) + index * sizeof(HelperEntry)), @result, sizeof(HelperEntry));
	end; (* GetNthElement *)

	procedure SetNthElement (entries: Handle; index: integer; var newValue: HelperEntry);
		(* This routine sets the index'th HelperEntry in the entries handle.
			0 <= index < number of entries
		*)
	begin
		BlockMoveData(@newValue, Ptr(longint(entries^) + index * sizeof(HelperEntry)), sizeof(HelperEntry));
	end; (* SetNthElement *)

	function EntryLocked (entries: Handle; index: integer): Boolean;
		(* This routine returns true if the index'th entry is locked. *)
		var
			tmpEntry: HelperEntry;
	begin
		GetNthElement(entries, index, tmpEntry);
		EntryLocked := tmpEntry.entryIsLocked;
	end; (* EntryLocked *)

	(* ***** Other Utilities for the Helpers Window ***** *)

	procedure DimButtons (wt: WindowType);
		(* Dims/undims the buttons in the dialog depending on the selection. This
			routine is typically called after the selection has been changed in some
			way.
		*)
		var
			window: DialogPtr;
			listH: ListHandle;
			entries: Handle;
			selection: integer;
			haveSelection: Boolean;
	begin
		window := gWindows[wt].window;
		listH := ListHandle(gWindows[wt].items[ditHelpersList]^.whatSpareData);
		entries := Handle(gWindows[wt].items[ditHelpersList]^.whatData);
		selection := LSelectedLine(listH);
		haveSelection := (selection <> -1);
		SetDControlEnable(window, ditHelpersAdd, not IsDocumentLocked);
		SetDControlEnable(window, ditHelpersDelete, haveSelection and not EntryLocked(entries, selection));
		SetDControlEnable(window, ditHelpersChange, haveSelection and not EntryLocked(entries, selection));
		DefaultButtonUserItem(window, ditHelpersDefaultUserItem);
	end; (* DimButtons *)

	function GetEntryName (list: ListHandle; listCell: Cell): Str255;
		(* Returns the name associated with the cell in the ListHandle.
			This is used by the LDoKey routine to support selection by
			typing into the list.
		*)
		var
			entries: Handle;
			entry: HelperEntry;
	begin
		GetEntryName := '';
		entries := Handle(gWindows[MapWindowPtrToWindowType(list^^.port)].items[ditHelpersList]^.whatData);
		GetNthElement(entries, listCell.v, entry);
		GetEntryName := concat(TPCopy(entry.urlScheme, length(kICHelper) + 1, 255), 
												GetIndStr(rMiscStrings, strLongDash), 
												entry.applicationForScheme.name);
	end; (* GetEntryName *)

	(* ***** Drawing ***** *)

	procedure HelperLDEF (message: integer; drawSelected: Boolean;
											var theCellRect: Rect;
											theCell: Cell;
											dataOffset: integer; datalen: integer;
											listH: ListHandle);
		(* This LDEF is responsible for drawing the main list item in the Helpers
			dialog. It basically is a simple dispatcher based on message type, where the
			only message types we're interested in are the draw and highlight message.
		*)

		procedure HelperLDEFDraw(entries: Handle);
			(* This is the core code for the LDEF, which draws a list cell.
				It's stupid for me to describe what the drawing looks like here.
				Run the application and take a look.
			*)
			var
				entry: HelperEntry;
				currentXOffset : integer;
				systemFontInfo: FontInfo;
				top: integer;
				iconRegion: RgnHandle;
		begin
			if datalen = 0 then begin
				(* Get the data associated with this entry. *)
				GetNthElement(entries, theCell.v, entry);
				
				(* Clean the cell to start with. *)
				EraseRect(theCellRect);

				(* Plot the icon, if we have the system software to do so. Maintain
					iconRegion as the region that was plotted, so we can exclude it
					from the highlighting. Also maintain the current X offset, so
					we can draw the text flush with the left hand side if we didn't
					plot an icon.
				*)
				currentXOffset := 0;
				iconRegion := NewRgn;
				if iconRegion <> nil then begin
					DrawListIcon('APPL', entry.applicationForScheme.fCreator,
											drawSelected, theCellRect,
											currentXOffset, iconRegion);
				end; (* if *)
				currentXOffset := currentXOffset + 2;

				(* Collect information about the fonts we're using. *)
				TextFont(systemFont);
				GetFontInfo(systemFontInfo);

				(* Draw the text. *)
				top := (kHelperCellHeight - (systemFontInfo.ascent + systemFontInfo.descent)) div 2;
				TextFont(systemFont);
				MoveTo(theCellRect.left + currentXOffset, theCellRect.top + top + systemFontInfo.ascent);
				DrawString(concat(TPCopy(entry.urlScheme, length(kICHelper) + 1, 255), GetIndStr(rMiscStrings, strLongDash), entry.applicationForScheme.name));

				(* If the cell is selected, draw the highlight region. *)
				if drawSelected then begin
					HighlightListCell(theCellRect, iconRegion);
				end; (* if *)
				if iconRegion <> nil then begin
					DisposeRgn(iconRegion);
				end; (* if *)
			end; (* if *)
		end; (* HelperLDEFDraw *)

		var
			entries: Handle;
	begin
		{$unused dataOffset}
		SetPort(listH^^.port);
		entries := Handle(gWindows[MapWindowPtrToWindowType(listH^^.port)].items[ditHelpersList]^.whatData);
		case message of
			lInitMsg: 
				;
			lDrawMsg,
			lHiliteMsg: 
				HelperLDEFDraw(entries);
			lCloseMsg: 
				;
			otherwise
				;
		end;
	end; (* HelperLDEF *)

	(* ***** Add Button ***** *)

	function DoAddButton(entries: Handle; lh: ListHandle): OSStatus;
		(* This routine is called in response to a request to add a new helper.
			Its basic purpose is to put up the Edit Helper modal dialog.
		*)
		var
			err: OSStatus;
			newEntry: HelperEntry;
			junkRow: integer;
	begin
		(* Create a new blank entry. *)
		newEntry.applicationForScheme.name := '';
		newEntry.entryIsLocked := false;
		newEntry.urlScheme := kICHelper;
		
		(* Put up a dialog to allow the user to edit it. *)
		err := DoEditHelperModalDialog(newEntry);
		
		(* If all is OK, add the entry to the entries data structure and the list. *)
		if err = noErr then begin
			err := PtrAndHand(@newEntry, entries, sizeof(newEntry));
			if err = noErr then begin
				junkRow := LAddRow(1, 32767, lh);
			end; (* if *)
		end; (* if *)
		
		DoAddButton := err;
	end; (* DoAddButton *)

	(* ***** Change Button ***** *)

	function DoChangeButton(itemToChange: integer; entries: Handle): OSStatus;
		(* This routine is called in response to a request to edit a helper.
			Its basic purpose is to put up the Edit Helper modal dialog.
		*)
		var
			err: OSStatus;
			entryToEdit: HelperEntry;
	begin
		(* Get the entry to edit, and pose the edit dialog. *)
		GetNthElement(entries, itemToChange, entryToEdit);
		err := DoEditHelperModalDialog(entryToEdit);
		
		(* If all is OK, update the entry in the entries data structure and the list. *)		
		if err = noErr then begin
			SetNthElement(entries, itemToChange, entryToEdit);
		end; (* if *)
		DoChangeButton := err;
	end; (* DoChangeButton *)

	(* ***** Delete Button ***** *)

	function DoDeleteButton (itemToDelete: integer; entries: Handle; listH: ListHandle): OSStatus;
		(* This routine is called in response to a request to delete a helper.
			It deletes the helper from the Internet Config database
			and our lists.
		*)
		var
			err: OSStatus;
			entryToDelete: HelperEntry;
			junkPos: longint;
			junk: OSStatus;
	begin
		(* First delete it out of Internet Config. *)
		err := ICBegin(WhatGetDocumentInstance, icReadWritePerm);
		if err = noErr then begin
			GetNthElement(entries, itemToDelete, entryToDelete);
			err := ICDeletePref(WhatGetDocumentInstance, entryToDelete.urlScheme);
			junk := ICEnd(WhatGetDocumentInstance);
		end; (* if *)
		
		(* Now delete it out of the entries handle and the list. *)
		if err = noErr then begin
			junkPos := Munger(entries, itemToDelete * sizeof(HelperEntry), nil, sizeof(HelperEntry), Ptr(-1), 0);
			err := MemError;
			if err = noErr then begin
				LDelRow(1, itemToDelete, listH);
			end; (* if *)
		end; (* if *)
		DoDeleteButton := err;
	end; (* DoDeleteButton *)

	(* ***** Layer Breaking Stuff ***** *)

	procedure HelperSetSelection(key : Str255);
		(* See comment in interface part. *)
		var
			junk : OSStatus;
			entryIndex:longint;
			window:WindowPtr;
			entries:Handle;
			entry:HelperEntry;
			listH : ListHandle;
	begin
		window := gWindows[WT_Helper].window;
		if window <> nil then begin
			(* Get some standard data structures. *)
			listH := ListHandle(gWindows[WT_Helper].items[ditHelpersList]^.whatSpareData);
			entries := Handle(gWindows[WT_Helper].items[ditHelpersList]^.whatData);
			
			(* Search for the key in the entries structure. *)
			key := concat(kICHelper, key);
			for entryIndex := 0 to GetHandleSize(entries) div sizeof(HelperEntry) - 1 do begin
				GetNthElement(entries, entryIndex, entry);
				if EqualString(entry.urlScheme, key, false, true) then begin

					(* Found it. Now select it and open the editor dialog. *)
					LSetSingleSelection(listH, entryIndex);
					DimButtons(WT_Helper);
					if GetDControlEnable(window, ditHelpersChange) then begin
						junk := DoChangeButton(entryIndex, entries);
					end; (* if *)
					leave;

				end; (* if *)
			end; (* for *)
		end; (* if *)
	end; (* HelperSetSelection *)

	(* ***** Standard What Interface ***** *)

	(* ***** What Init ***** *)

	function WhatInitHelper(dummy : integer) : OSStatus;
	begin
		{$unused dummy}
		gListUserItemUPP := NewUserItemProc(@ListUserItemUpdateProc);
		gFileMapLDEFUPP := NewListDefProc(@HelperLDEF);
		WhatInitHelper := noErr;
	end; (* WhatInitHelper *)

	(* ***** What Open ***** *)

	function WhatOpenHelper (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			err: OSStatus;
			window: DialogPtr;
			entries: Handle;
			attr: longint;
			numberOfEntries: longint;
			listItemRect: Rect;
			listDataRect: Rect;
			listCellSize: Point;
			listH: ListHandle;
			entryToAdd: HelperEntry;
			preferenceIndex: longint;
			size: longint;
	begin
		(* Prepare for failure. *)
		err := noErr;
		gWindows[wt].items[item]^.whatSpareData := nil;
		gWindows[wt].items[item]^.whatData := nil;
		
		(* Install the user item update proc for the list item. *)		
		window := gWindows[wt].window;
		SetDItemHandle(window, item, gListUserItemUPP);

		(* Create an empty list of entries. *)
		entries := NewHandle(0);
		if entries = nil then begin
			err := memFullErr;
		end; (* if *)
		
		(* Now build the list of entries by iterating through the IC preferences
			looking for all the preferences whose keys begin with "Helper".
		*)
		if err = noErr then begin
			preferenceIndex := 1;
			repeat
				err := ICGetIndPref(WhatGetDocumentInstance, preferenceIndex, entryToAdd.urlScheme);
				if err = noErr then begin
					if EqualString(TPCopy(entryToAdd.urlScheme, 1, length(kICHelper)), kICHelper, false, true) then begin
					
						(* We have a match, add it to the entries list. *)
						size := sizeof(entryToAdd.applicationForScheme);
						err := ICGetPref(WhatGetDocumentInstance, entryToAdd.urlScheme, attr, 
														@entryToAdd.applicationForScheme, size);
						if err = noErr then begin
							entryToAdd.entryIsLocked := btst(attr, ICattr_locked_bit) | IsDocumentLocked;
							err := PtrAndHand(@entryToAdd, entries, sizeof(HelperEntry));
						end; (* if *)
					end; (* if *)
				end; (* if *)
				preferenceIndex := preferenceIndex + 1;
			until err <> noErr;
		end; (* if *)
		if err = icPrefNotFoundErr then begin
			err := noErr;
		end; (* if *)
		
		(* Create the ListHandle for the list item. The list starts out one cell wide
			and numberOfEntries cells high. The horizontal cell size is governed by
			the width of the list user item and kHelperCellHeight. The total size of the list
			is determined by the list user item, minus the width of the scroll bar.
		*)
		if err = noErr then begin
			numberOfEntries := GetHandleSize(entries) div sizeof(HelperEntry);
			gWindows[wt].items[item]^.whatData := Ptr(entries);
			GetDItemRect(window, item, listItemRect);
			listItemRect.right := listItemRect.right - 15;
			SetRect(listDataRect, 0, 0, 1, numberOfEntries);
			SetPt(listCellSize, listItemRect.right - listItemRect.left, kHelperCellHeight);
			listH := LNew(listItemRect, listDataRect, listCellSize, rSimpleLDEF, window,
									true, false, false, true);
			if listH = nil then begin
				err := memFullErr;
			end; (* if *)
		end; (* if *)

		(* All is cool, finish up the open. *)
		if err = noErr then begin
			listH^^.refCon := ord(gFileMapLDEFUPP);
			listH^^.selFlags := lOnlyOne;
			gWindows[wt].items[item]^.whatSpareData := Ptr(listH);
			DimButtons(wt);
			SetupDefaultButtonUserItem(window, ditHelpersChange, ditHelpersDefaultUserItem);
			gWindows[wt].selectedItem := ditHelpersList;
		end; (* if *)

		WhatOpenHelper := err;
	end; (* WhatOpenHelper *)

	(* ***** What Flush ***** *)

	function WhatFlushHelper (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			err: OSStatus;
			firstErr: OSStatus;
			entries: Handle;
			entryIndex: integer;
			entryToSave: HelperEntry;
	begin
		{$unused item}
		(* Iterate through each of the entries, saving each to the Internet
			Config database.  Remember the first error that occurs, so we
			can continue saving but report that error to the user.
		*)
		err := noErr;
		entries := Handle(gWindows[wt].items[ditHelpersList]^.whatData);
		firstErr := noErr;
		for entryIndex := 0 to GetHandleSize(entries) div sizeof(HelperEntry) - 1 do begin
			GetNthElement(entries, entryIndex, entryToSave);
			if not entryToSave.entryIsLocked then begin
				err := ICSetPref(WhatGetDocumentInstance, entryToSave.urlScheme, ICattr_no_change, @entryToSave.applicationForScheme, sizeof(entryToSave.applicationForScheme));
			end; (* if *)
			if firstErr = noErr then begin
				firstErr := err;
			end; (* if *)
		end; (* for *)
		err := firstErr;
		WhatFlushHelper := err;
	end; (* WhatFlushHelper *)

	(* ***** What Close ***** *)

	function WhatCloseHelper (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
	begin
		if gWindows[wt].items[item]^.whatSpareData <> nil then begin
			LDispose(ListHandle(gWindows[wt].items[item]^.whatSpareData));
		end; (* if *)
		if Handle(gWindows[wt].items[ditHelpersList]^.whatData) <> nil then begin
			DisposeHandle(Handle(gWindows[wt].items[ditHelpersList]^.whatData));
		end; (* if *)
		WhatCloseHelper := noErr;
	end; (* WhatCloseHelper *)

	(* ***** What Key ***** *)

	function WhatKeyHelper (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			err: OSStatus;
			listH: ListHandle;
			entries: Handle;
	begin
		err := noErr;
		listH := ListHandle(gWindows[wt].items[item]^.whatSpareData);
		entries := Handle(gWindows[wt].items[ditHelpersList]^.whatData);
		if (event.what = keyDown) or (event.what = autoKey) then begin
			case chr(band(event.message, charCodeMask)) of
				kCRChar, kEnterChar:  begin
					if GetDControlEnable(gWindows[wt].window, ditHelpersChange) then begin
						FlashDItem(gWindows[wt].window, ditHelpersChange);
						err := DoChangeButton(LSelectedLine(listH), entries);
						if err = noErr then begin
							WhatDirtyDocument;
						end; (* if *)
					end; (* if *)
				end;
				otherwise begin
					LDoKey(listH, event, GetEntryName);
					DimButtons(wt);
				end;
			end; (* case *)
		end; (* if *)
		WhatKeyHelper := err;
	end; (* WhatKeyHelper *)

	(* ***** What Click ***** *)

	function WhatClickHelper (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			err: OSStatus;
			listH: ListHandle;
			entries: Handle;
			selectedListCell: integer;
	begin
		(* General preparation. *)
		entries := Handle(gWindows[wt].items[ditHelpersList]^.whatData);
		listH := ListHandle(gWindows[wt].items[ditHelpersList]^.whatSpareData);
		selectedListCell := LSelectedLine(listH);
		err := 1;

		(* Dispatch for each item in the dialog. *)
		case item of
			ditHelpersAdd: 
				err := DoAddButton(entries, listH);
			ditHelpersChange: 
				err := DoChangeButton(selectedListCell, entries);
			ditHelpersDelete: 
				if selectedListCell <> -1 then begin
					err := DoDeleteButton(selectedListCell, entries, listH);
				end; (* if *)
			ditHelpersList:  begin
				(* A click in the main list item. Call through to List Manager's LClick
					routine, and then attempt to open the Edit Helper dialog if
					it indicates a double click.
				*)
				GlobalToLocal(event.where);
				if LClick(event.where, 0, listH) then begin
					(* Call LSelectedLine again beacuse selection may have been changed by LClick. *)
					selectedListCell := LSelectedLine(listH);
					if EntryLocked(entries, selectedListCell) then begin
						WhatLockedAlert(wt, item);
						err := userCanceledErr;
					end else begin
						FlashDItem(gWindows[wt].window, ditHelpersChange);
						err := DoChangeButton(selectedListCell, entries);
					end; (* if *)
				end; (* if *)
			end;
		end; (* case *)

		(* Clean up. *)
		if err = noErr then begin
			WhatDirtyDocument;
		end; (* if *)
		if err = 1 then begin
			err := noErr;
		end; (* if *)
		DimButtons(wt);
		WhatClickHelper := err;
	end; (* WhatClickHelper *)

	(* ***** What Activate ***** *)

	function WhatActivateHelper (wt: WindowType; item: integer; activate: Boolean): OSStatus;
		(* See the comments in ICWindowGlobals. *)
	begin
		(* Pass the event straight through to List Manager. *)
		LActivate(activate, ListHandle(gWindows[wt].items[item]^.whatSpareData));
		WhatActivateHelper := noErr;
	end; (* WhatActivateHelper *)

	(* ***** What Cursor ***** *)

	function WhatCursorHelper (wt: WindowType; item: integer; pt: Point; cursorID: integer): OSStatus;
		(* See the comments in ICWindowGlobals and ICWhatSubs. *)
	begin
		WhatCursorHelper := WhatCursorListItem(wt, item, pt, cursorID);
	end; (* WhatCursorHelper *)
	
end. (* ICHelperWhat *)
