unit ICTextWhats;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the Text What. The Text What is weird because 
	ICWindows has incestuous knowledge of how it operates, and the two
	modules combine to implement all the necessary text behaviour.
	Putting all these interactions into the What API would have resulted
	in a substantially more complex interface, with very little gain for
	any Whats except the Text What, so I decided to just break a few layers.
	
	This module performs two basic functions:
	
	1. Loading and storing text using the IC API.  This is made substantially
		more difficult by all the different types of text that we want to
		edit, including single line strings, multiline text bodies, and
		scrambled passwords. The Text What determines what type
		of text it's editing using the various whatFlags.
		
	2. Editing text using TextEdit, as wrapped up by the ICText module.
	
	Text Whats store a ICText object of the text they're controlling in the
	whatData field of their WhatRecord. If the text is scrambled, they
	store a second ICText object (the "shadow") in the whatSpareData field
	so that they can store the bullets into the data and the real text into the
	spare data.
*)

interface

	uses
		ICWindowGlobals;

	(* ***** Standard What Interface ***** *)
	
	(* See the comments in ICWindowGlobals. *)
	
	function WhatInitText(dummy : integer) : OSStatus;
	function WhatOpenText (wt: WindowType; item: integer): OSStatus;
	function WhatFlushText (wt: WindowType; item: integer): OSStatus;
	function WhatCloseText (wt: WindowType; item: integer): OSStatus;
	function WhatKeyText (wt: WindowType; item: integer; event: EventRecord): OSStatus;
	function WhatClickText (wt: WindowType; item: integer; event: EventRecord): OSStatus;

	(* ***** Layer Breaking Interface ***** *)

	procedure AdjustTextMenu (wt: WindowType);
		(* Called by ICWindows to adjust the Edit menu, ie enable and disable
			items as determined by the current text selection.
		*)
	
	procedure DoTextMenu (wt: WindowType; menuItem: integer);
		(* Called by ICWindows in response to a menu selection on the Edit menu. *)
	
	procedure SelectTextItem (wt: WindowType; item: integer);
		(* This routine is called by ICWindows in response to the user
			tabbing through text fields, and also by ICEditPrefAppleEvents
			to select the preferences who key is the target of the event.
		*)

implementation

	uses
		Dialogs, 
		Scrap, 
		TextUtils, 
		Menus, 
		Fonts,

		InternetConfig, 
		
		ICResources, 
		ICStringLists,
		ICMiscSubs, 
		ICCommonSubs,
		ICText, 
		ICDialogs, 
		ICGlobals, 
		ICWindowGlobals;

	var
		gTextUserItemUPP : UserItemUPP;
		
	(* Text Whats recognise the following flags. These flags control how the
		string is stored, how it's displayed, and how it responds to user actions.
	*)
	const
		kWhatFlagsTextSingleLineBit = 0;
		(* No Returns in this text field. *)
		
		kWhatFlagsTextPStringBit = 1;
		(* Store this text as a PString as opposed to a text block. *)
		
		kWhatFlagsTextScrambledBit = 2;			
		(* Text is scrambled. Store it as a scrambled string and display it as bullets. *)
		
		kWhatFlagsTextMonospaceBit = 3;						
		(* Display text with monospaced font. *)
		
		kWhatFlagsTextFilterForDomainCharsBit = 4;		
		(* Filter out characters that aren't valid in a domain name,
			ie only allow kValidDomainCharacters to be typed into
			the field. Note that IC doesn't 100% guarantee
			that this text will only contain valid characters.
			You can still paste invalid characters in, or 
			write a little application that puts them in
			via the API.  This is more of a user interface
			hint than anything else.
		*)

		kValidDomainCharacters = [kSpaceChar, kTabChar, kCRChar,
													kBackSpaceChar,
													kLeftArrowChar, kRightArrowChar, kUpArrowChar, kDownArrowChar,
													'a'..'z', 'A'..'Z', '0'..'9',
													'.', '-',				(* as per Internet STD 13 *)
													':'						(* because some of these fields require a port specification *)
													];
		
		kWhatFlagsTextStringListBit = 5;						
		(* Store this text as a STR# as opposed to a text block. Strings are
			delimited by kWhitespaceCharacters.
		*)

		kWhitespaceCharacters = [' ', chr(9), chr(13)];

	(* ***** Drawing ***** *)
	
	procedure TextUserItemUpdateProc (window: DialogPtr; item: integer);
		(* The dialog item for a Text What is actually a user item.  The
			WhatOpenText installs TextUserItemUpdateProc as the update proc
			for this user item.  It simply calls through to ICText to
			do most of the work.
		*)
		var
			itemRect: Rect;
			savedWindowState : SavedWindowInfo;
	begin
		EnterWindow(window, 0, 0, [], savedWindowState);
		
		(* Draw the text. *)
		TextDraw(gWindows[MapWindowPtrToWindowType(window)].items[item]^.whatData);

		(* Draw the frame. *)
		GetDItemRect(window, item, itemRect);
		InsetRect(itemRect, -3, -3);
		PenNormal;
		FrameRect(itemRect);

		ExitWindow(savedWindowState);
	end; (* TextUserItemUpdateProc *)

	(* ***** Layer Breaking Stuff ***** *)

	function HaveTextSelection (wt: WindowType): Boolean;
		(* This routine returns true if the given window has a Text What
			with a selection in it.
		*)
		var
			startSelection : integer;
			endSelection: integer;
	begin
		HaveTextSelection := false;
		if GetSelectedTextItem(wt) > 0 then begin
			TextGetSelect(gWindows[wt].items[GetSelectedTextItem(wt)]^.whatData, startSelection, endSelection);
			HaveTextSelection := (startSelection <> endSelection);
		end; (* if *)
	end; (* HaveTextSelection *)

	procedure AdjustTextMenu (wt: WindowType);
		(* See comment in interface part.
		
			The Cut, Copy and Clear menu items are all controlled by the same
			algorithm. Basically the items are enabled if there's a selection
			and the preference isn't locked.
			
			Select All is enabled if a text item is the current focus and the
			entire text is not already selected.
			
			Paste is enabled if a text item is the current focus, the clipboard
			contains text and, if this is a single line text field, the text doesn't
			contain any CRs (except as the last character).  Pretty yucky.
		*)
		
		function ClipboardContainsText : Boolean;
			(* Returns true if the clipboard contains any TEXT data. *)
			var
				offset: longint;
		begin
			offset := 0;
			ClipboardContainsText := (GetScrap(nil, 'TEXT', offset) > 0);
		end; (* ClipboardContainsText *)
		
		function ClipboardContainsReturns : Boolean;
			(* Returns true if the clipboard contains any CRs, except
				if there's one CR at the end of the clipboard.
			*)
			var
				err: OSStatus;
				result: Boolean;
				offset: longint;
				textH: Handle;
				textSize: longint;
				textToSearchFor: integer;
				offsetOfFirstReturn: longint;
		begin
			result := false;			(* Default to saying we have no returns. *)
			
			(* Create an empty handle for the text. *)
			textH := NewHandle(0);
			err := MemError;
			if err = noErr then begin
			
				(* Fill out the handle from the scrap. *)
				offset := 0;
				textSize := GetScrap(textH, 'TEXT', offset);
				if textSize > 0 then begin
				
					(* Use Munger to search for a CR in the text. Note that we set
						textToSearchFor to $0D0D, which puts the ASCII for CR in both
						the high and low bytes of textToSearchFor. This is because
						Pascal is kinda stupid about how it handles single byte
						quantities, and this guarantees that we won't be stuffed
						up by this.
					*)
					textToSearchFor := $0D0D;
					offsetOfFirstReturn := Munger(textH, 0, @textToSearchFor, 1, nil, 0);
					
					if offsetOfFirstReturn < 0 then begin
						result := false;			(* No CRs in the text. *)
					end else if (offsetOfFirstReturn = textSize - 1) then begin
						(* There's a CR as the last byte of the text. Return true
							if that's is also the first byte.
						*)
						result := (offsetOfFirstReturn <= 0);
					end else begin
						(* CR in the middle of the text somewhere. *)
						result := true;
					end; (* if *)
				end; (* if *)
			end; (* if *)
			
			(* Clean up. *)
			if textH <> nil then begin
				DisposeHandle(textH);
			end; (* if *)
			if err <> noErr then begin
				result := true;
			end; (* if *)

			ClipboardContainsReturns := result;
		end; (* ClipboardContainsReturns *)
		
		var
			enableCut : Boolean;
			enableCopy : Boolean;
			enablePaste : Boolean;
			enableSelectAll: Boolean;
			editMenu: MenuHandle;
			selectedItem: integer;
			selectionStart : integer;
			selectionEnd: integer;
			textSize : longint;
	begin
		(* Default all the menu items to dimmed. *)
		enableCut := false;
		enableCopy := false;
		enablePaste := false;
		enableSelectAll := false;

		(* Find the select item and base the initial settings on the text selection
			in that item.
		*)
		selectedItem := GetSelectedTextItem(wt);
		if (wt <> WT_None) & (selectedItem > 0) then begin
			enableCut := HaveTextSelection(wt);
			enableCopy := HaveTextSelection(wt);
			enablePaste := ClipboardContainsText;
			TextGetSelect(gWindows[wt].items[selectedItem]^.whatData, selectionStart, selectionEnd);
			TextGetSize(gWindows[wt].items[selectedItem]^.whatData, textSize);
			enableSelectAll := (selectionStart <> 0) or (selectionEnd <> textSize);
		end; (* if *)
		
		(* Deal with the nasty problem of preventing the user pasting CRs into a single
			line field.
		*)
		if enablePaste & btst(gWindows[wt].items[selectedItem]^.whatFlags, kWhatFlagsTextSingleLineBit) then begin
			enablePaste := not ClipboardContainsReturns;
		end; (* if *)
		
		(* Disable Cut and Paste for locked items. *)
		if selectedItem > 0 then begin
			if WhatIsLocked(wt, selectedItem) then begin
				enablePaste := false;
				enableCut := false;
			end; (* if *)
		end; (* if *)

		(* Now enable/disable all the items. *)
		editMenu := GetMenuHandle(mEdit);
		SetMenuItemEnable(editMenu, iUndo, false);
		SetMenuItemEnable(editMenu, iCut, enableCut);
		SetMenuItemEnable(editMenu, iCopy, enableCopy);
		SetMenuItemEnable(editMenu, iPaste, enablePaste);
		SetMenuItemEnable(editMenu, iClear, enableCut);
		SetMenuItemEnable(editMenu, iSelectAll, enableSelectAll);
	end; (* AdjustTextMenu *)

	procedure DoTextMenu (wt: WindowType; menuItem: integer);
		(* See comment in interface part. *)

		procedure RemoveReturns(textH : Handle);
			(* This routine removes the final CR from a handle that is
				about to be pasted into a single line field. It actually
				removes all of the text after that return as well, but 
				we should never hit that case because the menu item
				would be disabled.
			*)
			var
				textToSearchFor: integer;
				offsetOfFirstReturn: longint;
				textSize: longint;
		begin
			textSize := GetHandleSize(textH);
			if textSize > 0 then begin
				textToSearchFor := $0D0D;
				offsetOfFirstReturn := Munger(textH, 0, @textToSearchFor, 1, nil, 0);
				if offsetOfFirstReturn > 0 then begin
					SetHandleSize(textH, offsetOfFirstReturn);
				end; (* if *)
			end; (* if *)
		end; (* RemoveReturns *)

		procedure DoPaste(what : WhatRecordPtr);
			(* Paste the clipboard into the specified What.  This is harder than it
				seems.  We first have to get rid of CRs in text that's pasted into
				single line fields.  Then we have to handle pasting text into
				scrambled fields.
			*)
			var
				textH : Handle;
				textSize : longint;
				offset : longint;
		begin
			textH := NewHandle(0);
			if textH <> nil then begin
				offset := 0;
				textSize := GetScrap(textH, 'TEXT', offset);
				if textSize > 0 then begin
					(* It it's a single line field, remove any CRs. *)
					if btst(what^.whatFlags, kWhatFlagsTextSingleLineBit) then begin
						RemoveReturns(textH);
					end; (* if *)

					(* If it's a scrambled field, paste the real data into the shadow field,
						then munge the textH to hold all bullets and fall through into
						the standard paste code.
					*)
					textSize := GetHandleSize(textH);
					if btst(what^.whatFlags, kWhatFlagsTextScrambledBit) then begin
						TextClear(what^.whatSpareData);
						TextInsert(what^.whatSpareData, textH);
						BlockFill (textH^, textSize, ord(gBulletChar));
					end; (* if *)
					
					TextClear(what^.whatData);
					TextInsert(what^.whatData, textH);
				end; (* if *)
				DisposeHandle(textH);
			end; (* if *)
		end; (* DoPaste *)

		var
			data: Ptr;
			spareData: Ptr;
			selectedItem: integer;
			textIsScrambled : Boolean;
			selStart, selEnd : integer;
	begin
		(* selectedItem should never be -1 because otherwise the menu commands
			wouldn't have been enabled.
		*)
		selectedItem := GetSelectedTextItem(wt);
		data := gWindows[wt].items[selectedItem]^.whatData;
		spareData := gWindows[wt].items[selectedItem]^.whatSpareData;

		(* Handle the common case when the text item is locked. *)
		if (menuItem in [iCut, iPaste, iClear]) & WhatIsLocked(wt, selectedItem) then begin
			WhatLockedAlert(wt, selectedItem);
			menuItem := 0;		(* Prevent the command from being executed. *)
		end; (* if *)

		textIsScrambled := btst(gWindows[wt].items[selectedItem]^.whatFlags, kWhatFlagsTextScrambledBit);
		if textIsScrambled then begin
			(* Synchronise the shadow field's selection with the
				visible field's one.  We need to do this because the
				user can change the selection in the main field
				without us noticing.  When the Edit menu commands
				get executed, we have to make sure that the selections
				are synched otherwise they'll do different things
				in the shadow and visible fields.
			*)
			TextGetSelect(data, selStart, selEnd);
			TextSetSelect(spareData, selStart, selEnd);
		end; (* if *)
		
		(* Now dispatch the commands.  Commands that modify
			the text must also modify the shadow text field if there
			is one.
		*)
		case menuItem of
			iCut:
				begin
					if textIsScrambled then begin
						TextCut(spareData);
					end; (* if *)
					TextCut(data);
				end;
			iCopy:
				TextCopy(data);
			iClear:
				begin
					if textIsScrambled then begin
						TextClear(spareData);
					end; (* if *)
					TextClear(data);
				end;
			iSelectAll:
				TextSetSelect(data, 0, 32767);
			iPaste: 
				DoPaste(gWindows[wt].items[selectedItem]);
			otherwise
				;
		end; (* case *)
	end; (* DoTextMenu *)

	procedure SwitchToTextItem (wt: WindowType; item: integer);
		(* This internal routine sets item to be the currently selected item
			for the designated window as far as the keyboard focus and
			highlighting is concerned.  It doesn't actually select the text in
			the item.
		*)
	begin
		(* If there was a previously selected item, deactivate it. *)
		if GetSelectedTextItem(wt) > 0 then begin
			TextActivate(gWindows[wt].items[GetSelectedTextItem(wt)]^.whatData, false);
		end; (* if *)
		
		(* Activate the new item and mark it to receive keyboard events. *)
		TextActivate(gWindows[wt].items[item]^.whatData, (FrontWindow = gWindows[wt].window) & InForeground);
		gWindows[wt].selectedItem := item;
	end; (* SwitchToTextItem *)

	procedure SelectTextItem (wt: WindowType; item: integer);
		(* See comment in interface part. *)
	begin
		SwitchToTextItem(wt, item);
		TextSetSelect(gWindows[wt].items[item]^.whatData, 0, 32767);
	end; (* SelectTextItem *)

	(* ***** Utility Routines ***** *)
	
	procedure ScrambleHandle (textH: Handle);
		(* This routine scrambles textH according to the algorithm dictated
			in the IC Programming Docs. This is just to obfuscate idle password
			snooping.
			
			This transfer is identpotent, ie if you apply it twice, you get back
			to where you started.
		*)
		var
			textIndex: longint;
			tmptext: BigBufferPtr;
	begin
		tmptext := BigBufferPtr(textH^);
		for textIndex := 1 to GetHandleSize(textH) do begin
			tmptext^[textIndex - 1] := band(bxor(tmptext^[textIndex - 1], $55 + textIndex), $FF);
		end; (* for *)
	end; (* ScrambleHandle *)
	
	function FixCyberdogStrH(strH : Handle) : Boolean;
		(* Attempt to fix a Cyberdog 1.0 STR# resource.  Cyberdog 1.0 writes the
			resource out with a 1 byte count instead of the correct two byte count.  
			IC attempts to fix this by adding a single 0 byte at the front.  If the 
			resulting STR# is valid, then we've fixed the STR# and this routine returns
			true.  If not, this routine returns false.

			Cyberdog 1.1 corrects this bug, by the way.
		*)
		var
			result : Boolean;
			junk : longint;
			zeroData : integer;
	begin
		result := (strH <> nil) & (strH^ <> nil) & (GetHandleSize(strH) >= 1);
		if result then begin
			(* Use Munger to insert a zero byte at the start of the handle. *)
			zeroData := 0;
			junk := Munger(strH, 0, nil, 0, @zeroData, 1);
			result := ValidStrList(strH);
		end; (* if *)
		FixCyberdogStrH := result;
	end; (* FixCyberdogStrH *)

	procedure UnpackStringList(var textH : Handle);
		(* Unpack the STR# list that's currently in textH and
			replace textH with a handle to the text that's contained
			in those strings (separated by spaces).
		*)
		var
			stringList : Handle;
			stringIndex : integer;
			tempString : Str255;
			junk : OSStatus;
	begin
		stringList := textH;
		textH := NewHandle(0);
		if textH <> nil then begin

			(* Walk through the string list, adding each string to the text. *)
			for stringIndex := 1 to CountStrList(stringList) do begin
				tempString := GetIndStrList(stringList, stringIndex);
				if stringIndex <> CountStrList(stringList) then begin
					tempString := concat(tempString, ' ');
				end; (* if *)
				junk := PtrAndHand(@tempString[1], textH, length(tempString));
			end; (* for *)
			
			DisposeHandle(stringList);
		end else begin
			(* Whoops, failed to allocate the space for the new data,
				put the old handle back.
			*)
			textH := stringList;
		end; (* if *)
	end; (* UnpackStringList *)

	procedure PackStringList(var textH : Handle);
		(* Parse the text contained in textH, break it into words delimited
			by kWhitespaceCharacters, and add each word to a string list.
			When done, replace textH with the newly created string list.
			
			This isn't an efficient implementation, but it sure was easy to code.
		*)
		var
			stringList : Handle;
			charIndex : longint;
			tempString : Str255;
	begin
		stringList := NewStrList;
		if stringList <> nil then begin
			
			(* Keep doing this until we get to the end of the text. *)
			charIndex := 0;
			while charIndex < GetHandleSize(textH) do begin
				
				(* Accumulate non-whitespace characters in tempString. *)
				tempString := '';
				while (charIndex < GetHandleSize(textH)) & not (BigCharArrayHandle(textH)^^[charIndex] in kWhitespaceCharacters) do begin
					tempString := concat(tempString, BigCharArrayHandle(textH)^^[charIndex]);
					charIndex := charIndex + 1;
				end; (* while *)
				
				(* Skip over whitespace characters. *)
				while (charIndex < GetHandleSize(textH)) & (BigCharArrayHandle(textH)^^[charIndex] in kWhitespaceCharacters) do begin
					charIndex := charIndex + 1;
				end; (* while *)
				
				(* If we found a word, put it at the end of the string list. *)
				if tempString <> '' then begin
					InsertIndStrList (stringList, CountStrList(stringList) + 1, tempString);
				end; (* if *)
			end; (* while *)
			
			(* Clean up. *)
			DisposeHandle(textH);
			textH := stringList;
		end; (* if *)
	end; (* PackStringList *)

	function LocalTextCreate (var data: univ Ptr; wt: WindowType; item: integer): OSStatus;
		var
			font : integer;
			size : integer;
	begin
		if btst(gWindows[wt].items[item]^.whatFlags, kWhatFlagsTextMonospaceBit) then begin
			GetFNum("Monaco", font);
			size := 9;
		end else begin
			font := systemFont;
			size := 12;
		end; (* if *)

		LocalTextCreate := TextCreate(data, gWindows[wt].window, item, font, size, WhatIsLocked(wt, item));
	end; (* LocalTextCreate *)

	(* ***** Standard What Interface ***** *)

	(* ***** What Init ***** *)

	function WhatInitText(dummy : integer) : OSStatus;
	begin
		{$unused dummy}
		gTextUserItemUPP := NewUserItemProc(@TextUserItemUpdateProc);
		WhatInitText := noErr;
	end; (* WhatInitText *)

	(* ***** What Open ***** *)

	function WhatOpenText (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals.
		
			There is a lot of subtlety in this routine. For example, we carefully
			place newly created data structures into the global
			gWindows[wt].items[item]^.xxx as we create them, so that if
			the creation fails at any stage the close routine will get
			a chance to clean them up.
			
			Also the ordering of some lines is important in a non-obvious way.
			For example, you have to unscramble the text after unpacking
			it from the save format, otherwise you find yourself
			scrambling length bytes and such.
		*)

		function CreateShadowText(textH: Handle) : OSStatus;
			(* This routine moves the existing text data (currently
				residing in whatSpareData) off the screen,
				and creates a new text field, full of bullets, on screen
				for the user to see and edit.
			*)
			var
				err : OSStatus;
				itemRect : Rect;
		begin
			(* Move the existing item off the screen. *)
			GetDItemRect(gWindows[wt].window, item, itemRect);
			OffsetRect(itemRect, 16000, 0);
			TextMove(gWindows[wt].items[item]^.whatSpareData, itemRect);
			
			(* Fill textH with bullets. *)
			BlockFill(textH^, GetHandleSize(textH), ord(gBulletChar));
			
			(* And create a second, on screen, text item that we're going to use
				for editing.
			*)
			err := LocalTextCreate(gWindows[wt].items[item]^.whatData, wt, item);
			if err = noErr then begin
				TextInsert(gWindows[wt].items[item]^.whatData, textH);
			end; (* if *)
			CreateShadowText := err;
		end; (* CreateShadowText *)

		var
			err: OSStatus;
			attr: longint;
			textH: Handle;
			junk: longint;
			flags: longint;
			textIsPString: Boolean;
			textIsScrambled: Boolean;
			textIsStringList : Boolean;
	begin
		(* Nil out the global data pointers early to ensure that we can clean up safely. *)
		gWindows[wt].items[item]^.whatData := nil;
		gWindows[wt].items[item]^.whatSpareData := nil;

		(* Prepare for failure. *)		
		textH := nil;
		
		(* Create some local Booleans that represent the whatFlags. *)
		flags := gWindows[wt].items[item]^.whatFlags;
		textIsPString := btst(flags, kWhatFlagsTextPStringBit);
		textIsStringList := btst(flags, kWhatFlagsTextStringListBit);
		textIsScrambled := btst(flags, kWhatFlagsTextScrambledBit);

		(* Get the text associated with this key. It's important to realise that
			ICGetPrefHandle will return an empty handle if the key does not exist.
		*)
		err := ICGetPrefHandle(WhatGetDocumentInstance, gWindows[wt].items[item]^.whatKey, attr, textH);

		(* Now start the long process of creating this item... *)
		
		if err = noErr then begin
			WhatProcessAttributes(wt, item, attr);

			(* Deal with the various formats that the handle might take. *)
			
			if textIsPString then begin
				(* If it's a PString, simply remove the first byte off the front. *)
				junk := Munger(textH, 0, nil, 1, @junk, 0);
			end else if textIsStringList then begin
				(* If it's a STR#, there's two stages.  The first to to accomodate
					for the badly formatted STR#'s output by Cyberdog 1.0.
				*)
				if not ValidStrList(textH) then begin
					if not FixCyberdogStrH(textH) then begin
						(* Couldn't fix it, use an empty list instead. *)
						SetHandleSize(textH, 2);
						IntegerPtr(textH^)^ := 0;
					end; (* if *)
				end; (* if *)
				
				(* The next step is to walk the STR# and create the text
					handle from the string list.
				*)
				UnpackStringList(textH);
			end; (* if *)
			
			(* If the text is scrambled, unscramble it. It's important that this be
				done after the PString unpacked, lest we try to unscramble
				the length byte.
			*)
			if textIsScrambled then begin
				ScrambleHandle(textH);
			end; (* if *)

			(* Create the TextEdit data structures associated with this item. *)
			err := LocalTextCreate(gWindows[wt].items[item]^.whatData, wt, item);
		end; (* if *)
		
		(* Now install the user item update proc, and jam the text into TextEdit. *)
		if err = noErr then begin
			SetDItemHandle(gWindows[wt].window, item, gTextUserItemUPP);
			TextInsert(gWindows[wt].items[item]^.whatData, textH);
		end; (* if *)
		
		(* If the item is scrambled, create an offscreen shadow text data
			structure for it.
		*)
		if (err = noErr) & textIsScrambled then begin
			gWindows[wt].items[item]^.whatSpareData := gWindows[wt].items[item]^.whatData;
			gWindows[wt].items[item]^.whatData := nil;
			err := CreateShadowText(textH);
		end; (* if *)
		
		(* Clean up. *)
		if textH <> nil then begin
			DisposeHandle(textH);
		end; (* if *)
		
		WhatOpenText := err;
	end; (* WhatOpenText *)

	(* ***** What Flush ***** *)

	function WhatFlushText (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals.
		
			Again, this routine is surprisingly tricky. The basic process is:
			
			1. Get the old text.
			2. Normalise it based on the formatting flags.
			3. Get the current text.
			4. Normalise it based on the formatting flags.
			5. Compare the two.
			6. If they're different, write the new text out.
			
			We do this because it's very hard to keep of what events can possibly
			modify a text field, so we cop out and just do a bulk comparison
			at the end.  Of course, as with many cop outs, this may have prooved
			to be harder than doing it the right way.  But this approach does
			have other advantages, namely that data isn't written unless it's
			actually been changed.
		*)
		var
			err: OSStatus;
			textH: Handle;
			oldTextH: Handle;
			attr: longint;
			junk: longint;
			stringLengthToInsert: integer;
			flags: longint;
			textIsPString: Boolean;
			textIsScrambled: Boolean;
			textIsStringList : Boolean;
	begin
		(* Prepare for failure. *)
		textH := nil;
		oldTextH := nil;

		(* Create some local Booleans that represent the whatFlags. *)
		flags := gWindows[wt].items[item]^.whatFlags;
		textIsPString := btst(flags, kWhatFlagsTextPStringBit);
		textIsStringList := btst(flags, kWhatFlagsTextStringListBit);
		textIsScrambled := btst(flags, kWhatFlagsTextScrambledBit);
		
		(* Get the old text, ie the text that's currently stored in the preference.
			Remember that ICGetPrefHandle will return an empty handle if the key
			does not exist, so if this is the first time we've been flushed, we have
			to be prepared to deal with an empty handle coming back.
		*)
		err := ICGetPrefHandle(WhatGetDocumentInstance, gWindows[wt].items[item]^.whatKey, attr, oldTextH);
		
		(* Normalise an empty text handle so it looks like the empty version
			of whatever we're expecting.
		*)
		if (err = noErr) & (GetHandleSize(oldTextH) = 0) then begin
			if textIsPString then begin
				(* Create an empty PString. *)
				SetHandleSize(oldTextH, 1);
				err := MemError;
				if err = noErr then begin
					oldTextH^^ := 0;
				end; (* if *)
			end else if textIsStringList then begin
				(* Create an empty string list. *)
				SetHandleSize(oldTextH, sizeof(integer));
				err := MemError;
				if err = noErr then begin
					IntegerPtr(oldTextH^)^ := 0;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		(* Now extract the current text from the current TextEdit data
			structures.  This is a surprisingly complicated process...
		*)
		if err = noErr then begin
			textH := NewHandle(0);
			err := MemError;
		end; (* if *)
		if err = noErr then begin
			(* If the text is scrambled, get the real text from the whatSpareData
				rather than the bullets (which are in the whatData field.
			*)
			if textIsScrambled then begin
				TextGet(gWindows[wt].items[item]^.whatSpareData, textH);
			end else begin
				TextGet(gWindows[wt].items[item]^.whatData, textH);
			end; (* if *)
			
			(* Now scramble the text if necessary. *)
			if textIsScrambled then begin
				ScrambleHandle(textH);
			end; (* if *)
			
			(* Now format it for saving. *)
			if textIsPString then begin
				(* PStrings shall be less than 255 chars long. *)
				if GetHandleSize(textH) > 255 then begin
					SetHandleSize(textH, 255);
				end; (* if *)
				(* Use Munger to insert the length byte at the front of the PString.
					Pascal is kinda stupid about how it handles single byte
					quantities, so we put the length in both bytes of a two byte
					integer by multiplying it by $0101.
				*)
				stringLengthToInsert := GetHandleSize(textH) * $0101;
				junk := Munger(textH, 0, nil, 0, @stringLengthToInsert, 1);
			end else if textIsStringList then begin
				PackStringList(textH);
			end; (* if *)
			
			(* Now compare the textH to the oldTextH and write the data back if
				the text has been changed.
			*) 
			if not WhatIsLocked(wt, item) & 
						((GetHandleSize(textH) <> GetHandleSize(oldTextH)) 
							| not BlockCompare(textH^, oldTextH^, GetHandleSize(textH))) then begin
				err := ICSetPrefHandle(WhatGetDocumentInstance, gWindows[wt].items[item]^.whatKey, ICattr_no_change, textH);
				if err = noErr then begin
					WhatDirtyDocument;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		(* Clean up. *)
		if textH <> nil then begin
			DisposeHandle(textH);
		end; (* if *)
		if oldTextH <> nil then begin
			DisposeHandle(oldTextH);
		end; (* if *)
		WhatFlushText := err;
	end; (* WhatFlushText *)

	(* ***** What Close ***** *)

	function WhatCloseText (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals.

			Note that this routine relies on the fact that TextDestroy
			handles a nil input just fine.
		*)
	begin
		(* Destroy the main editable text. *)
		TextDestroy(gWindows[wt].items[item]^.whatData);
		(* If the text is scrambled, also destroy the shadow text. *)
		if btst(gWindows[wt].items[item]^.whatFlags, kWhatFlagsTextScrambledBit) then begin
			TextDestroy(gWindows[wt].items[item]^.whatSpareData);
		end; (* if *)
		WhatCloseText := noErr;
	end; (* WhatCloseText *)

	(* ***** What Key ***** *)

	function WhatKeyText (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			flags: longint;
			textIsScrambled : Boolean;
			textIsSingleLine : Boolean;
			textIsFilteredForDomains : Boolean;
			typedChar : char;
			selectionStart : integer;
			selectionEnd: integer;
	begin
		(* Create some local Booleans that represent the whatFlags. *)
		flags := gWindows[wt].items[item]^.whatFlags;
		textIsScrambled := btst(flags, kWhatFlagsTextScrambledBit);
		textIsSingleLine := btst(flags, kWhatFlagsTextSingleLineBit);
		textIsFilteredForDomains := btst(flags, kWhatFlagsTextFilterForDomainCharsBit);
		
		(* Get the ASCII value typed. *)
		typedChar := chr(band(event.message, charCodeMask));
		
		(* First do a quick check to make sure the characters have a hope of being
			valid.
		*)
		if (textIsSingleLine & (typedChar = chr(13))) |
					(textIsScrambled & ((typedChar <> chr(8)) & (typedChar < ' '))) |
					(textIsFilteredForDomains & not (typedChar in kValidDomainCharacters)) then begin
			SysBeep(10);
		end else begin
		
			(* Now prevent typing in locked fields. *)
			if WhatIsLocked(wt, item) & DirtyKey(typedChar) then begin
				WhatLockedAlert(wt, item);
			end else begin
			
				if textIsScrambled then begin
					(* If the text is scrambled, use the standard algorithm to support
						shadowed text:

						1. Get the selection in the visible field.
						2. Set that selection in the shadow field.
						3. Type the character in the shadow field.
						4. Change the character to a bullet and type it in the visible field.
					*)
					TextGetSelect(gWindows[wt].items[item]^.whatData, selectionStart, selectionEnd);
					TextSetSelect(gWindows[wt].items[item]^.whatSpareData, selectionStart, selectionEnd);
					TextKey(gWindows[wt].items[item]^.whatSpareData, event);
					if (typedChar <> chr(8)) & DirtyKey(typedChar) then begin
						event.message := ord(gBulletChar);
					end; (* if *)
					TextKey(gWindows[wt].items[item]^.whatData, event);
					
				end else begin

					(* Not a scrambled field, just pass the character along. *)
					TextKey(gWindows[wt].items[item]^.whatData, event);
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		WhatKeyText := noErr;
	end; (* WhatKeyText *)

	(* ***** What Click ***** *)

	function WhatClickText (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* See the comments in ICWindowGlobals. *)
	begin
		(* Unitlaterally switch to the text item. While this might seem to
			involve a lot of extra work if we're currently in this item, this
			work just isn't noticable.
		*)
		SwitchToTextItem(wt, item);

		TextClick(gWindows[wt].items[item]^.whatData, event);
		WhatClickText := noErr;
	end; (* WhatClickText *)

end. (* TextWhats *)